##############################################################################
# MIT License
#
# Copyright (c) 2021 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

from typing import Any

from dash import html
from dash_svg import G, Path, Rect, Svg, Text

from utils import schema
from utils.logger import console_error
from utils.utils import format_scientific_notation_if_needed

# Constants for display formatting
DEFAULT_MAX_LENGTH = 6
DEFAULT_PRECISION = 1
DEFAULT_SCIENTIFIC_WIDTH = 8


def insert_chart_data(mem_data: list[dict[str, Any]], base_data: schema.Workload) -> G:
    if len(mem_data) != 1:
        console_error("Memory Chart config doesn't follow expected formatting")

    table_config = mem_data[0]["metric_table"]
    original_df = base_data.dfs[table_config["id"]]
    display_columns = original_df.columns.values.tolist().copy()
    display_df = original_df[display_columns]

    alias = display_df["Metric"].values
    values = display_df["Value"].values

    memchart_values: dict[str, Any] = {}
    for i in range(0, len(alias)):
        memchart_values[alias[i]] = values[i]

    return G(
        className="data",
        children=[
            # ----------------------------------------
            # Instr Buff Block
            # TODO: double check wave_occupancy
            Text(
                x="52",
                y="313",
                id="wave_occ",
                fill="#FFFF33",
                fontSize="20px",
                fontWeight="bold",
                children=format_value_for_display(
                    memchart_values.get("Wavefront Occupancy")
                ),
            ),
            Text(
                x="49",
                y="394",
                id="wave_life",
                fill="#FFFF33",
                fontSize="20px",
                fontWeight="bold",
                children=format_value_for_display(memchart_values.get("Wave Life")),
            ),
            # ----------------------------------------
            # Instr Dispatch Block
            Text(
                x="386",
                y="46",
                id="salu",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("SALU")),
            ),
            Text(
                x="386",
                y="96",
                id="smem",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("SMEM")),
            ),
            Text(
                x="386",
                y="146",
                id="valu",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VALU")),
            ),
            Text(
                x="386",
                y="196",
                id="mfma",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("MFMA")),
            ),
            Text(
                x="386",
                y="245",
                id="vmem",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VMEM")),
            ),
            Text(
                x="386",
                y="296",
                id="lds",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("LDS")),
            ),
            Text(
                x="386",
                y="344",
                id="gws",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("GWS")),
            ),
            Text(
                x="386",
                y="396",
                id="br",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("BR")),
            ),
            # ----------------------------------------
            # Exec Block
            Text(
                x="480",
                y="99",
                id="active_cu",
                fill="#FFFF33",
                fontSize="20px",
                fontWeight="bold",
                children=format_value_for_display(memchart_values.get("Active CUs")),
            ),  # x=454
            Text(
                x="580",
                y="154",
                id="vgpr",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VGPR")),
            ),
            Text(
                x="581",
                y="183",
                id="sgpr",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("SGPR")),
            ),
            Text(
                x="580",
                y="226",
                id="lds_alloc",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(
                    memchart_values.get("LDS Allocation")
                ),
            ),
            Text(
                x="580",
                y="255",
                id="scratch_alloc",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(
                    memchart_values.get("Scratch Allocation")
                ),
            ),
            Text(
                x="580",
                y="298",
                id="wavefronts",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Wavefronts")),
            ),
            Text(
                x="580",
                y="328",
                id="workgroups",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Workgroups")),
            ),
            # ----------------------------------------
            # LDS Block
            Text(
                x="723",
                y="78",
                id="lds_req",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("LDS Req")),
            ),
            Text(
                x="839",
                y="85",
                id="lds_util",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("LDS Util")),
            ),
            Text(
                x="839",
                y="117",
                id="lds_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("LDS Latency")),
            ),
            # ----------------------------------------
            # Vector L1 Cache Block
            Text(
                x="708",
                y="204",
                id="vl1_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Rd")),
            ),
            Text(
                x="708",
                y="233",
                id="vl1_wr",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Wr")),
            ),
            Text(
                x="716",
                y="265",
                id="vl1_atom",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Atomic")),
            ),
            Text(
                x="840",
                y="193",
                id="vl1_hit",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Hit")),
            ),
            Text(
                x="840",
                y="224",
                id="vl1_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Lat")),
            ),
            Text(
                x="840",
                y="256",
                id="vl1_coales",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Coalesce")),
            ),
            Text(
                x="838",
                y="288",
                id="vl1_stall",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1 Stall")),
            ),
            Text(
                x="1000",
                y="203",
                id="vl1_l2_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1_L2 Rd")),
            ),
            Text(
                x="1000",
                y="232",
                id="vl1_l2_wr",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1_L2 Wr")),
            ),
            Text(
                x="1008",
                y="264",
                id="vl1_l2_atom",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("VL1_L2 Atomic")),
            ),
            # ----------------------------------------
            # Scalar L1D Cache Block
            Text(
                x="709",
                y="384",
                id="sl1_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("sL1D Rd")),
            ),
            Text(
                x="838",
                y="372",
                id="sl1_hit",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("sL1D Hit")),
            ),
            Text(
                x="838",
                y="404",
                id="sl1_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("sL1D Lat")),
            ),
            Text(
                x="1000",
                y="351",
                id="sl1_l2_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("sL1D_L2 Rd")),
            ),
            Text(
                x="1000",
                y="380",
                id="sl1_l2_wr",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("sL1D_L2 Wr")),
            ),
            Text(
                x="1008",
                y="412",
                id="sl1_l2_atom",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(
                    memchart_values.get("sL1D_L2 Atomic")
                ),
            ),
            # ----------------------------------------
            # Instr L1  Cache Block
            Text(
                x="492",
                y="498",
                id="il1_fetch",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("IL1 Fetch")),
            ),
            Text(
                x="837",
                y="491",
                id="il1_hit",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("IL1 Hit")),
            ),
            Text(
                x="837",
                y="522",
                id="il1_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("IL1 Lat")),
            ),
            Text(
                x="1015",
                y="500",
                id="il1_l2_req",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("IL1_L2 Rd")),
            ),
            # ----------------------------------------
            # L2 Cache Block(inside)
            Text(
                x="1145",
                y="213",
                id="l2_rd",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Rd")),
            ),
            Text(
                x="1145",
                y="238",
                id="l2_wr",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Wr")),
            ),
            Text(
                x="1145",
                y="264",
                id="l2_atom",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Atomic")),
            ),
            Text(
                x="1145",
                y="292",
                id="l2_hit",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Hit")),
            ),
            Text(
                x="1145",
                y="356",
                id="l2_rd_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Rd Lat")),
            ),
            Text(
                x="1145",
                y="382",
                id="l2_wr_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("L2 Wr Lat")),
            ),
            # ----------------------------------------
            # Fabric Block
            Text(
                x="1317",
                y="243",
                id="l2_fabric_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Fabric_L2 Rd")),
            ),
            Text(
                x="1317",
                y="272",
                id="l2_fabric_wr",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Fabric_L2 Wr")),
            ),
            Text(
                x="1319",
                y="303",
                id="l2_fabric_atom",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(
                    memchart_values.get("Fabric_L2 Atomic")
                ),
            ),
            Text(
                x="1435",
                y="285",
                id="fabric_rd_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Fabric Rd Lat")),
            ),
            Text(
                x="1435",
                y="310",
                id="fabric_wr_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("Fabric Wr Lat")),
            ),
            Text(
                x="1435",
                y="336",
                id="fabric_atom_lat",
                fill="rgb(0, 0, 0)",
                fontSize="12px",
                children=format_value_for_display(
                    memchart_values.get("Fabric Atomic Lat")
                ),
            ),
            Text(
                x="1578",
                y="240",
                id="hbm_rd",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("HBM Rd")),
            ),
            Text(
                x="1577",
                y="269",
                id="hbm_wr",
                fill="#FFFFFF",
                fontSize="12px",
                children=format_value_for_display(memchart_values.get("HBM Wr")),
            ),
        ],
    )


def get_memchart(
    mem_data: list[dict[str, Any]], base_data: schema.Workload
) -> html.Section:
    return html.Section(
        id="memchart",
        children=[
            html.Div(
                id="memchart-svg",
                children=[
                    Svg(
                        children=[
                            G(
                                children=[
                                    G(
                                        className="instr-buff",
                                        children=[
                                            Rect(x="30", y="25.5"),
                                            Rect(x="20", y="32"),
                                            Rect(x="10", y="42"),
                                            Rect(x="0", y="52"),
                                        ],
                                    ),
                                    G(
                                        className="fabric",
                                        children=[
                                            Rect(x="1373", y="213"),
                                            Rect(x="1363", y="203"),
                                        ],
                                    ),
                                    G(
                                        className="cache",
                                        children=[
                                            Rect(x="757", y="460"),
                                            Rect(x="757", y="345"),
                                            Rect(id="a3", x="757", y="32"),
                                            Rect(id="b3", x="757", y="165"),
                                        ],
                                    ),
                                    G(
                                        className="fabric-connections",
                                        children=[
                                            Rect(x="1383", y="56"),
                                            Rect(x="1383", y="451"),
                                            Rect(x="1606.69", y="227.43"),
                                        ],
                                    ),
                                    G(
                                        className="inner-inst-buff",
                                        children=[
                                            Rect(x="20", y="92"),
                                            Rect(x="20", y="170.28"),
                                        ],
                                    ),
                                    G(
                                        className="misc-rec",
                                        children=[
                                            Rect(x="1063", y="32"),
                                            Rect(id="a6", x="326.25", y="138.25"),
                                            Rect(id="b6", x="450", y="122"),
                                        ],
                                    ),
                                    G(
                                        className="val-1",
                                        children=[
                                            Rect(x="319", y="32", rx="3", ry="3"),
                                            Rect(x="319", y="82", rx="3", ry="3"),
                                            Rect(x="319", y="132", rx="3", ry="3"),
                                            Rect(x="319", y="182", rx="3", ry="3"),
                                            Rect(x="319", y="231", rx="3", ry="3"),
                                            Rect(x="319", y="282", rx="3", ry="3"),
                                            Rect(x="319", y="329.5", rx="3", ry="3"),
                                            Rect(x="319", y="382", rx="3", ry="3"),
                                            Rect(x="1367.69", y="271", rx="3", ry="3"),
                                            Rect(
                                                x="1367.69", y="296.5", rx="3", ry="3"
                                            ),
                                            Rect(
                                                x="1367.69", y="322.5", rx="3", ry="3"
                                            ),
                                            Rect(x="1078", y="199", rx="3", ry="3"),
                                            Rect(x="1078", y="224.5", rx="3", ry="3"),
                                            Rect(x="1078", y="250.5", rx="3", ry="3"),
                                            Rect(x="771.44", y="103", rx="3", ry="3"),
                                            Rect(
                                                x="770.44", y="358.75", rx="3", ry="3"
                                            ),
                                            Rect(
                                                x="770.44", y="390.25", rx="3", ry="3"
                                            ),
                                            Rect(x="769.44", y="477", rx="3", ry="3"),
                                            Rect(x="769.44", y="508.5", rx="3", ry="3"),
                                            Rect(x="1078", y="278", rx="3", ry="3"),
                                            Rect(x="1078", y="342.5", rx="3", ry="3"),
                                            Rect(x="1078", y="368.5", rx="3", ry="3"),
                                            Rect(x="772.44", y="179", rx="3", ry="3"),
                                            Rect(
                                                x="772.44", y="210.18", rx="3", ry="3"
                                            ),
                                            Rect(x="771.44", y="71.28", rx="3", ry="3"),
                                            Rect(x="772.44", y="242", rx="3", ry="3"),
                                            Rect(x="770.44", y="274.5", rx="3", ry="3"),
                                        ],
                                    ),
                                    G(
                                        className="val-2",
                                        children=[
                                            Rect(x="362", y="32", rx="3", ry="3"),
                                            Rect(x="362", y="82", rx="3", ry="3"),
                                            Rect(x="362", y="132", rx="3", ry="3"),
                                            Rect(x="362", y="182", rx="3", ry="3"),
                                            Rect(x="362", y="231", rx="3", ry="3"),
                                            Rect(x="362", y="282", rx="3", ry="3"),
                                            Rect(x="362", y="329.5", rx="3", ry="3"),
                                            Rect(x="362", y="382", rx="3", ry="3"),
                                        ],
                                    ),
                                    G(
                                        className="val-3",
                                        children=[
                                            Rect(x="1410.69", y="271", rx="3", ry="3"),
                                            Rect(
                                                x="1410.69", y="296.5", rx="3", ry="3"
                                            ),
                                            Rect(
                                                x="1410.69", y="322.5", rx="3", ry="3"
                                            ),
                                            Rect(x="1121", y="199", rx="3", ry="3"),
                                            Rect(x="1121", y="224.5", rx="3", ry="3"),
                                            Rect(x="1121", y="250.5", rx="3", ry="3"),
                                            Rect(x="814.44", y="103", rx="3", ry="3"),
                                            Rect(
                                                x="813.44", y="358.75", rx="3", ry="3"
                                            ),
                                            Rect(
                                                x="813.44", y="390.25", rx="3", ry="3"
                                            ),
                                            Rect(x="812.44", y="477", rx="3", ry="3"),
                                            Rect(x="812.44", y="508.5", rx="3", ry="3"),
                                            Rect(x="1121", y="278", rx="3", ry="3"),
                                            Rect(x="1121", y="342.5", rx="3", ry="3"),
                                            Rect(x="1121", y="368.5", rx="3", ry="3"),
                                            Rect(x="815.44", y="179", rx="3", ry="3"),
                                            Rect(
                                                x="815.44", y="210.18", rx="3", ry="3"
                                            ),
                                            Rect(x="814.44", y="71.28", rx="3", ry="3"),
                                            Rect(x="815.44", y="242", rx="3", ry="3"),
                                            Rect(x="813.44", y="274.5", rx="3", ry="3"),
                                        ],
                                    ),
                                    G(
                                        className="val-4",
                                        children=[
                                            Rect(x="460", y="212.5", rx="3", ry="3"),
                                            Rect(x="460", y="241", rx="3", ry="3"),
                                            Rect(x="460", y="284.54", rx="3", ry="3"),
                                            Rect(x="460", y="314", rx="3", ry="3"),
                                            Rect(x="460", y="140.32", rx="3", ry="3"),
                                            Rect(x="460", y="169.16", rx="3", ry="3"),
                                        ],
                                    ),
                                    G(
                                        className="val-5",
                                        children=[
                                            Rect(
                                                x="548.25",
                                                y="212.98",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                            Rect(
                                                x="548.25",
                                                y="241.48",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                            Rect(
                                                x="548.25",
                                                y="285.02",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                            Rect(
                                                x="548.25",
                                                y="314.48",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                            Rect(
                                                x="548.25",
                                                y="140.8",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                            Rect(
                                                x="549.5",
                                                y="169.64",
                                                rx="2.86",
                                                ry="2.86",
                                            ),
                                        ],
                                    ),
                                    G(
                                        className="lines-arrows",
                                        children=[
                                            Path(
                                                id="p1",
                                                d="M 100 243.72 L 120 220.28 L "
                                                "475 220.28 L 495 243.72 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 62 L 285.99 62.15",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 112 L 285.15 111.92",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 162 L 285.57 161.69",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 212 L 285.15 211.85",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 285.66 262.41 L 260 262.07",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 312 L 284.73 312.18",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 362 L 284.28 361.95",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 260 412 L 285.57 412.12",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 310.02 62.15 L 413.63 62.01",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 62 L 411.89 65.51 L "
                                                "413.63 62.01 L 411.88 58.51 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.92 111.92 L 413.63 112",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 112 L 411.88 115.49 L "
                                                "413.63 112 L 411.88 108.49 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.08 162.08 L 413.63 162",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 162 L 411.88 165.51 L "
                                                "413.63 162 L 411.88 158.51 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 310 212 L 413.63 212",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 212 L 411.88 215.5 "
                                                "L 413.63 212 L 411.88 208.5 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.92 262.02 L 413.63 262",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 262 L 411.88 265.5 L "
                                                "413.63 262 L 411.88 258.5 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.36 312.18 L 413.63 312.01",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 312 L 411.89 315.51 L "
                                                "413.63 312.01 L 411.88 308.51 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.08 361.95 L 413.63 362",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 362 L 411.88 365.5 L "
                                                "413.63 362 L 411.88 358.5 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 309.36 408.56 L 413.63 408.97",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 418.88 409 L 411.87 412.47 L "
                                                "413.63 408.97 L 411.9 405.47 Z",
                                                fill="#ff8000",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 750 207 L 652.37 207",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 647.12 207 L 654.12 203.5 L "
                                                "652.37 207 L 654.12 210.5 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 646 236.57 L 743.63 236.03",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 748.88 236.01 L 741.9 239.54 L "
                                                "743.63 236.03 L 741.86 232.54 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 750 502 L 110 502 Q 100 "
                                                "502 100 492 L 100 468.37",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 100 463.12 L 103.5 470.12 L "
                                                "100 468.37 L 96.5 470.12 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1050 504 L 942.37 504",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 937.12 504 L 944.12 500.5 L "
                                                "942.37 504 L 944.12 507.5 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1597.69 242.93 L 1534.06 242.93",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1528.81 242.93 L 1535.81 239.43 L "
                                                "1534.06 242.93 L 1535.81 246.43 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1527.69 274.64 L 1591.32 274.64",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1596.57 274.64 L 1589.57 278.14 L "
                                                "1591.32 274.64 L 1589.57 271.14 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1413 196 L 1413 132.37",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1413 127.12 L 1416.5 134.12 L "
                                                "1413 132.37 L 1409.5 134.12 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1443 126 L 1443 189.63",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1443 194.88 L 1439.5 187.88 L "
                                                "1443 189.63 L 1446.5 187.88 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1413.36 441 L 1413.03 377.37",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1413.01 372.12 L 1416.54 379.1 L "
                                                "1413.03 377.37 L 1409.54 379.14 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1443 371 L 1443.33 432.13",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1443.35 437.38 L 1439.82 430.4 L "
                                                "1443.33 432.13 L 1446.82 430.36 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1145.25 341.38 L 1141.75 334.38 L "
                                                "1145.25 336.13 L 1148.75 334.38 Z",
                                                fill="rgb(0, 0, 0)",
                                                stroke="rgb(0, 0, 0)",
                                            ),
                                            Path(
                                                d="M 740 82 L 652.37 82",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 647.12 82 L 654.12 78.5 L 652.37 "
                                                "82 L 654.12 85.5 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 744 386.75 L 656.37 386.75",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 651.12 386.75 L 658.12 383.25 L "
                                                "656.37 386.75 L 658.12 390.25 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 651.37 269 L 743.63 269",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 646.12 269 L 653.12 265.5 L "
                                                "651.37 269 L 653.12 272.5 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 748.88 269 L 741.88 272.5 L "
                                                "743.63 269 L 741.88 265.5 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1042 206.41 L 944.37 206.41",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 939.12 206.41 L 946.12 202.91 L "
                                                "944.37 206.41 L 946.12 209.91 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 938 235.98 L 1035.63 235.44",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1040.88 235.42 L 1033.9 238.95 L "
                                                "1035.63 235.44 L 1033.86 231.95 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 943.37 268.41 L 1035.63 268.41",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 938.12 268.41 L 945.12 264.91 L "
                                                "943.37 268.41 L 945.12 271.91 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1040.88 268.41 L 1033.88 271.91 L "
                                                "1035.63 268.41 L 1033.88 264.91 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1042 354.32 L 944.37 354.32",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 939.12 354.32 L 946.12 350.82 L "
                                                "944.37 354.32 L 946.12 357.82 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 938 383.89 L 1035.63 383.35",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1040.88 383.33 L 1033.9 386.86 L "
                                                "1035.63 383.35 L 1033.86 379.86 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 943.37 416.32 L 1035.63 416.32",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 938.12 416.32 L 945.12 412.82 L "
                                                "943.37 416.32 L 945.12 419.82 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1040.88 416.32 L 1033.88 419.82 L "
                                                "1035.63 416.32 L 1033.88 412.82 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1355 245.75 L 1257.37 245.75",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1252.12 245.75 L 1259.12 242.25 L "
                                                "1257.37 245.75 L 1259.12 249.25 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1251 275.32 L 1348.63 274.78",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1353.88 274.76 L 1346.9 278.29 L "
                                                "1348.63 274.78 L 1346.86 271.29 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1256.37 307.75 L 1348.63 307.75",
                                                fill="none",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1251.12 307.75 L 1258.12 304.25 L "
                                                "1256.37 307.75 L 1258.12 311.25 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                d="M 1353.88 307.75 L 1346.88 311.25 L "
                                                "1348.63 307.75 L 1346.88 304.25 Z",
                                                fill="#00cccc",
                                                stroke="#00cccc",
                                            ),
                                            Path(
                                                id="p2",
                                                d="M 235 67 L 245 57 L "
                                                "265 57 L 275 67 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 56 L 250 56",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 65 L 250 65",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 74 L 250 74",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 47.5 L 250 47.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p3",
                                                d="M 235 117 L 245 107 L 265 "
                                                "107 L 275 117 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 106 L 250 106",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 115 L 250 115",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 124 L 250 124",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 97.5 L 250 97.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p4",
                                                d="M 235 167 L 245 157 L "
                                                "265 157 L 275 167 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 156 L 250 156",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 165 L 250 165",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 174 L 250 174",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 147.5 L 250 147.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p5",
                                                d="M 235 217 L 245 207 L "
                                                "265 207 L 275 217 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 206 L 250 206",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 215 L 250 215",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 224 L 250 224",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 197.5 L 250 197.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p6",
                                                d="M 235 267 L 245 257 L "
                                                "265 257 L 275 267 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 256 L 250 256",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 265 L 250 265",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 274 L 250 274",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 247.5 L 250 247.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p7",
                                                d="M 235 317 L 245 307 L "
                                                "265 307 L 275 317 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 306 L 250 306",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 315 L 250 315",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 324 L 250 324",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 297.5 L 250 297.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p8",
                                                d="M 235 367 L 245 357 L "
                                                "265 357 L 275 367 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 356 L 250 356",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 365 L 250 365",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 374 L 250 374",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 230 347.5 L 250 347.5",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                id="p9",
                                                d="M 235 417 L 245 407 L "
                                                "265 407 L 275 417 Z",
                                                fill="#ffffff",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 220 406 L 250 406",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 210 415 L 250 415",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                            Path(
                                                d="M 200 424 L 250 424",
                                                fill="none",
                                                stroke="#ff8000",
                                            ),
                                        ],
                                    ),
                                    G(
                                        className="labels",
                                        children=[
                                            Text(
                                                x="12",
                                                y="278",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Wave Occupancy",
                                            ),
                                            Text(
                                                x="12",
                                                y="363",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Wave Life",
                                            ),
                                            Text(
                                                x="1428",
                                                y="80",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="xGMI /",
                                            ),
                                            Text(
                                                x="1428",
                                                y="105",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="PCIe",
                                            ),
                                            Text(
                                                x="1428",
                                                y="487",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="GMI",
                                            ),
                                            Text(
                                                x="1652",
                                                y="263",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="HBM",
                                            ),
                                            Text(
                                                x="1438",
                                                y="230",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="Fabric",
                                            ),
                                            Text(
                                                x="360",
                                                y="47",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="SALU:",
                                            ),
                                            Text(
                                                x="360",
                                                y="97",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="SMEM:",
                                            ),
                                            Text(
                                                x="360",
                                                y="147",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="VALU:",
                                            ),
                                            Text(
                                                x="360",
                                                y="197",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="MFMA:",
                                            ),
                                            Text(
                                                x="360",
                                                y="246",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="VMEM:",
                                            ),
                                            Text(
                                                x="360",
                                                y="297",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="LDS:",
                                            ),
                                            Text(
                                                x="360",
                                                y="344",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="GWS:",
                                            ),
                                            Text(
                                                x="360",
                                                y="397",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Br:",
                                            ),
                                            Text(
                                                x="1463",
                                                y="285",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1408",
                                                y="286",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="1463",
                                                y="310",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1408",
                                                y="311",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="1463",
                                                y="336",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1408",
                                                y="337",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                            Text(
                                                x="1118",
                                                y="214",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="1118",
                                                y="239",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="1118",
                                                y="265",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                            Text(
                                                x="867",
                                                y="117",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="812",
                                                y="117",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Lat:",
                                            ),
                                            Text(
                                                x="866",
                                                y="372",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="810",
                                                y="373",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Hit:",
                                            ),
                                            Text(
                                                x="866",
                                                y="404",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="810",
                                                y="405",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Lat:",
                                            ),
                                            Text(
                                                x="865",
                                                y="491",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="809",
                                                y="492",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Hit:",
                                            ),
                                            Text(
                                                x="865",
                                                y="522",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="809",
                                                y="523",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Lat:",
                                            ),
                                            Text(
                                                x="1556",
                                                y="239",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="1554",
                                                y="269",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="699",
                                                y="77",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Req:",
                                            ),
                                            Text(
                                                x="684",
                                                y="204",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="684",
                                                y="233",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="696",
                                                y="265",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                            Text(
                                                x="102",
                                                y="312",
                                                fill="#FFFF33",
                                                fontSize="20px",
                                                fontWeight="bold",
                                                children="per-GCD",
                                            ),
                                            Text(
                                                x="102",
                                                y="393",
                                                fill="#FFFF33",
                                                fontSize="20px",
                                                fontWeight="bold",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1173",
                                                y="292",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="1118",
                                                y="293",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Hit:",
                                            ),
                                            Text(
                                                x="1173",
                                                y="356",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1118",
                                                y="357",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="1173",
                                                y="382",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="1118",
                                                y="383",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="32",
                                                y="126",
                                                fill="#FFFFFF",
                                                fontSize="14px",
                                                children="Wave 0 Instr buff",
                                            ),
                                            Text(
                                                x="32",
                                                y="205",
                                                fill="#FFFFFF",
                                                fontSize="14px",
                                                children="Wave N-1 Instr buff",
                                            ),
                                            Text(
                                                x="442",
                                                y="69",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Active CUs",
                                            ),
                                            Text(
                                                x="868",
                                                y="193",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="812",
                                                y="194",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Hit:",
                                            ),
                                            Text(
                                                x="868",
                                                y="224",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="cycles",
                                            ),
                                            Text(
                                                x="812",
                                                y="225",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Lat:",
                                            ),
                                            Text(
                                                x="867",
                                                y="85",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="812",
                                                y="85",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Util:",
                                            ),
                                            Text(
                                                x="868",
                                                y="256",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="813",
                                                y="256",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Coales:",
                                            ),
                                            Text(
                                                x="432",
                                                y="18",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Exec",
                                            ),
                                            Text(
                                                x="12",
                                                y="18",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Instr Buff",
                                            ),
                                            Text(
                                                x="250",
                                                y="18",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Instr Dispatch",
                                            ),
                                            Text(
                                                x="761",
                                                y="26",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="LDS",
                                            ),
                                            Text(
                                                x="760",
                                                y="158",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Vector L1 Cache",
                                            ),
                                            Text(
                                                x="761",
                                                y="337",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Scalar L1D Cache",
                                            ),
                                            Text(
                                                x="761",
                                                y="451",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                children="Instr L1 Cache",
                                            ),
                                            Text(
                                                x="1153",
                                                y="63",
                                                fill="#FFFFFF",
                                                fontSize="20px",
                                                textAnchor="middle",
                                                children="L2 Cache",
                                            ),
                                            Text(
                                                x="991",
                                                y="499",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Req:",
                                            ),
                                            Text(
                                                x="866",
                                                y="288",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                children="%",
                                            ),
                                            Text(
                                                x="811",
                                                y="288",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Stall:",
                                            ),
                                            Text(
                                                x="468",
                                                y="497",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Fetch:",
                                            ),
                                            Text(
                                                x="1153",
                                                y="333",
                                                fill="#FFFFFF",
                                                fontSize="14px",
                                                textAnchor="middle",
                                                textDecoration="underline",
                                                children="Latency",
                                            ),
                                            Text(
                                                x="543",
                                                y="227",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="LDS Alloc:",
                                            ),
                                            Text(
                                                x="543",
                                                y="255",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Scratch Alloc:",
                                            ),
                                            Text(
                                                x="543",
                                                y="299",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wavefronts:",
                                            ),
                                            Text(
                                                x="543",
                                                y="328",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Workgroups:",
                                            ),
                                            Text(
                                                x="543",
                                                y="155",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="VGPRs:",
                                            ),
                                            Text(
                                                x="544",
                                                y="183",
                                                fill="rgb(0, 0, 0)",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="SGPRs:",
                                            ),
                                            Text(
                                                x="684",
                                                y="384",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="976",
                                                y="204",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="976",
                                                y="232",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="988",
                                                y="264",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                            Text(
                                                x="976",
                                                y="352",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="976",
                                                y="380",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="988",
                                                y="412",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                            Text(
                                                x="1292",
                                                y="243",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Rd:",
                                            ),
                                            Text(
                                                x="1293",
                                                y="272",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Wr:",
                                            ),
                                            Text(
                                                x="1301",
                                                y="304",
                                                fill="#FFFFFF",
                                                fontSize="12px",
                                                textAnchor="end",
                                                children="Atomic:",
                                            ),
                                        ],
                                    ),
                                    insert_chart_data(mem_data, base_data),
                                ]
                            )
                        ],
                        viewBox="-0.5 -0.5 1698 543",
                    )
                ],
            )
        ],
    )


def format_value_for_display(value: Any, max_length: int = DEFAULT_MAX_LENGTH) -> str:  # noqa: ANN401
    """
    Format a value (int, float, or str) into a concise string suitable for display.

    The function attempts to convert strings to numeric types if possible.
    It then decides between normal decimal notation and scientific notation
    based on length constraints and value magnitude.

    If the formatted string is too long, it truncates it gracefully,
    preserving scientific notation exponent parts where applicable.

    Parameters:
    - value: The input value to format. Can be int, float, or string.
             Strings representing numbers are converted to numeric types if possible.
    - max_length: Maximum allowed length of the output string.
                  Longer strings are truncated with an ellipsis ('…').

    Returns:
    - A string representation of the input value, formatted either
      in fixed-point or scientific notation, and truncated if too long.
      Returns "N/A" if the value is invalid (e.g., None or NaN).
    """

    if value is None:
        return "N/A"

    if isinstance(value, str):
        try:
            if "." in value:
                # when dot is in the string, we know it's a
                # float number and convert with "float"
                value = float(value)
            else:
                # without dot, we assume it's an integer and convert with "int"
                value = int(value)
        except ValueError:
            # when conversion fails, the string is neither legit float or int,
            # then assume it's invalid and display "N/A"
            return "N/A"

    if isinstance(value, (int, float)):
        is_negative = value < 0
        abs_val = abs(value)

        if isinstance(abs_val, float):
            if value != value:
                return "N/A"

            if abs_val.is_integer():
                normal = str(int(abs_val))
            else:
                normal = f"{abs_val:.1f}"
        else:
            normal = str(abs_val)

        sci = format_scientific_notation_if_needed(
            abs_val,
            align=">",
            width_align=DEFAULT_SCIENTIFIC_WIDTH,
            precision=DEFAULT_PRECISION,
            fmt_type_align="e",
            max_length=max_length,
        ).strip()

        # Choose shorter notation or if normal too long
        if len(sci) < len(normal) or len(normal) > max_length:
            value = sci
        else:
            value = normal

        if is_negative:
            value = f"-{value}"

    else:
        value = str(value)

    # Custom truncation logic:
    if len(value) > max_length:
        if "e" in value.lower():
            e_index = value.lower().index("e")
            mantissa = value[:e_index]
            exponent = value[e_index:]
            max_mantissa_len = max_length - len(exponent)
            if max_mantissa_len < 1:
                value = f"{exponent[: max_length - 1]}…"
            else:
                truncated_mantissa = mantissa[:max_mantissa_len]
                value = truncated_mantissa + exponent
        else:
            value = f"{value[: max_length - 1]}…"

    return value



##############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Label, ListItem, ListView

from rocprof_compute_tui.widgets.instant_button import InstantButton


class ClickableListItem(ListItem):
    """A ListItem that highlights on single click without selecting."""

    def on_mouse_down(self, event) -> None:  # noqa: ANN001
        event.stop()

        # Find parent ListView
        parent = self.parent
        list_view = None
        while parent is not None:
            if isinstance(parent, ListView):
                list_view = parent
                break
            parent = parent.parent

        if list_view is None:
            return

        # Determine index
        try:
            index = list_view.children.index(self)
        except ValueError:
            return

        # Update highlight immediately
        list_view.index = index
        list_view.refresh(layout=True)
        self.refresh(layout=True)

    def on_click(self, event) -> None:  # noqa: ANN001
        event.stop()

    def on_mouse_up(self, event) -> None:  # noqa: ANN001
        event.stop()


class RecentDirectoriesScreen(ModalScreen):
    """Modal screen to display recent directories as clickable list items."""

    def __init__(self, recent_dirs: list[str], current_dir: str | None = None) -> None:
        super().__init__()
        self.recent_dirs = recent_dirs
        self.current_dir = current_dir

    # -------------------------------------------------------------------------
    # Compose UI
    # -------------------------------------------------------------------------
    def compose(self) -> ComposeResult:
        with Container(id="recent-modal"):
            yield Label("Recent Directories", id="recent-title")

            if self.recent_dirs:
                with ListView(id="recent-list"):
                    for i, directory in enumerate(self.recent_dirs):
                        # Normal-looking list item (NO InstantButton inside)
                        yield ClickableListItem(
                            Label(directory),
                            id=f"recent-row-{i}",
                        )
            else:
                yield Label("No recent directories found", id="no-recent")

            with Horizontal(id="recent-buttons"):
                yield InstantButton("Select", variant="primary", id="select-recent")
                yield InstantButton("Close", variant="default", id="close-recent")

    # -------------------------------------------------------------------------
    # Initialize highlight when screen opens
    # -------------------------------------------------------------------------
    def on_mount(self) -> None:
        if not self.recent_dirs:
            return

        list_view = self.query_one("#recent-list", ListView)

        # Default to first item
        index = 0

        # If current_dir matches one of the recent dirs
        if self.current_dir:
            try:
                index = self.recent_dirs.index(self.current_dir)
            except ValueError:
                pass

        list_view.index = index
        list_view.focus()

    # -------------------------------------------------------------------------
    # Buttons
    # -------------------------------------------------------------------------
    def on_instant_button_instant_pressed(
        self, event: InstantButton.InstantPressed
    ) -> None:
        """Handle Select / Close / row-selection."""
        button = event.button
        bid = button.id

        # Close
        if bid == "close-recent":
            event.stop()
            self.dismiss(None)
            return

        # Select highlighted row
        if bid == "select-recent":
            event.stop()
            list_view = self.query_one("#recent-list", ListView)
            idx = list_view.index or 0
            self.dismiss(self.recent_dirs[idx])
            return

    # -------------------------------------------------------------------------
    # Keyboard activation (Enter or double-click)
    # -------------------------------------------------------------------------
    def on_list_view_selected(self, event: ListView.Selected) -> None:
        idx = event.list_view.index or 0
        self.dismiss(self.recent_dirs[idx])



##############################################################################
# MIT License
#
# Copyright (c) 2021 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

import logging
import os
import sys
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar

R = TypeVar("R")

# Define the colors
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[%dm"

COLORS = {
    "WARNING": YELLOW,
    "INFO": GREEN,
    "DEBUG": BLUE,
    "CRITICAL": RED,
    "ERROR": RED,
    "TRACE": MAGENTA,
}

# Constants
TRACE_LEVEL = logging.DEBUG - 5

LOG_LEVEL_MAPPING = {
    "DEBUG": logging.DEBUG,
    "debug": logging.DEBUG,
    "TRACE": TRACE_LEVEL,
    "trace": TRACE_LEVEL,
    "INFO": logging.INFO,
    "info": logging.INFO,
    "ERROR": logging.ERROR,
    "error": logging.ERROR,
}


def demarcate(function: Callable[..., R]) -> Callable[..., R]:
    def wrap_function(*args: Any, **kwargs: Any) -> R:
        trace_logger(f"----- [entering function] -> {function.__qualname__}()")
        result = function(*args, **kwargs)
        trace_logger(f"----- [exiting  function] -> {function.__qualname__}()")
        return result

    return wrap_function


def console_error(*argv: Any, exit: bool = True) -> None:
    if len(argv) > 1:
        logging.error(f"[{argv[0]}] {argv[1]}")
    elif len(argv) == 1:
        logging.error(f"{argv[0]}")
    else:
        logging.error("Empty error message")
    if exit:
        sys.exit(1)


def console_log(*argv: Any, indent_level: int = 0) -> None:
    indent = ""
    if indent_level >= 1:
        indent = " " * (3 * indent_level) + "|-> "  # spaces per indent level

    if len(argv) > 1:
        logging.info(indent + f"[{argv[0]}] {argv[1]}")
    elif len(argv) == 1:
        logging.info(indent + f"{argv[0]}")
    else:
        logging.info(indent + "Empty log message")


def console_debug(*argv: Any) -> None:
    if len(argv) > 1:
        logging.debug(f"[{argv[0]}] {argv[1]}")
    elif len(argv) == 1:
        logging.debug(f"{argv[0]}")
    else:
        logging.debug("Empty debug message")


def console_warning(*argv: Any) -> None:
    if len(argv) > 1:
        logging.warning(f"[{argv[0]}] {argv[1]}")
    elif len(argv) == 1:
        logging.warning(f"{argv[0]}")
    else:
        logging.warning("Empty warning message")


def trace_logger(message: str, *args: Any, **kwargs: Any) -> None:
    logging.log(TRACE_LEVEL, message, *args, **kwargs)


# Define the formatter
class ColoredFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        levelname = record.levelname
        if levelname in COLORS:
            levelname_color = (
                COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            )
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


class ColoredFormatterAll(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        levelname = record.levelname
        if levelname in COLORS:
            if levelname == "INFO":
                log_fmt = "%(message)s"
            else:
                log_fmt = (
                    f"{COLOR_SEQ % (30 + COLORS[levelname])}"
                    f"%(levelname)s: %(message)s{RESET_SEQ}"
                )
            formatter = logging.Formatter(log_fmt)
            return formatter.format(record)
        return super().format(record)


class PlainFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        if record.levelno == logging.ERROR:
            self._style._fmt = "%(levelname)s %(message)s"
        else:
            self._style._fmt = "%(message)s"
        return logging.Formatter.format(self, record)


# Setup console handler - provided as separate function to be called
# prior to argument parsing
def setup_console_handler() -> None:
    logging.getLogger().handlers.clear()
    # register a trace level logger
    logging.addLevelName(TRACE_LEVEL, "TRACE")
    setattr(logging, "TRACE", TRACE_LEVEL)
    setattr(logging, "trace", trace_logger)

    color_setting = 1
    if "ROCPROFCOMPUTE_COLOR" in os.environ.keys():
        color_setting = int(os.environ["ROCPROFCOMPUTE_COLOR"])

    if color_setting == 0:
        # non-colored
        formatter = PlainFormatter()
    elif color_setting == 1:
        # colored loglevel and with non-colored message
        formatter = ColoredFormatter("%(levelname)16s %(message)s")
    elif color_setting == 2:
        # non-colored levelname included
        formatter = logging.Formatter("%(levelname)5s %(message)s")
    elif color_setting == 3:
        # no color or levelname for INFO, other log messages entirely in color
        formatter = ColoredFormatterAll()
    else:
        print("Unsupported setting for ROCPROFCOMPUTE_COLOR - set to 0, 1, 2 or 3.")
        sys.exit(1)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    console_handler.set_name("console")
    logging.getLogger().addHandler(console_handler)


# Setup file handler - enabled in profile mode
def setup_file_handler(loglevel: int, workload_dir: str) -> None:
    filename = str(Path(workload_dir) / "log.txt")
    file_handler = logging.FileHandler(filename, "w")
    file_loglevel = min([loglevel, logging.INFO])
    file_handler.setLevel(file_loglevel)
    file_handler.setFormatter(logging.Formatter("%(message)s"))
    logging.getLogger().addHandler(file_handler)


# Setup logger priority - called after argument parsing
def setup_logging_priority(
    verbosity: int, quietmode: bool, appmode: str, guimode: Optional[bool] = None
) -> int:
    # set loglevel based on selected verbosity and quietmode
    levels = [logging.INFO, logging.DEBUG, TRACE_LEVEL]

    if quietmode:
        loglevel = logging.ERROR
    else:
        loglevel = levels[min(verbosity, len(levels) - 1)]  # cap to last level index

    # optional: suppress Werkzeug's messages in analyze GUIs.
    if quietmode and "analyze" in appmode and guimode:
        werkzeug_logger = logging.getLogger("werkzeug")
        werkzeug_logger.setLevel(logging.ERROR)

    # optional: override of default loglevel via env variable which takes precedence
    if "ROCPROFCOMPUTE_LOGLEVEL" in os.environ.keys():
        loglevel = os.environ["ROCPROFCOMPUTE_LOGLEVEL"]

        if loglevel in LOG_LEVEL_MAPPING:
            loglevel = LOG_LEVEL_MAPPING[loglevel]
        else:
            print(f"Ignoring unsupported ROCPROFCOMPUTE_LOGLEVEL setting ({loglevel})")
            sys.exit(1)

    # update console loglevel based on command-line args/env settings
    for handler in logging.getLogger().handlers:
        if handler.get_name() == "console":
            handler.setLevel(loglevel)

    # set global loglevel to min of console/file settings in profile mode
    if appmode == "profile":
        global_loglevel = min([logging.INFO, loglevel])
        logging.getLogger().setLevel(global_loglevel)
    else:
        logging.getLogger().setLevel(loglevel)

    return loglevel



##############################################################################
# MIT License
#
# Copyright (c) 2021 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################
"""Get host/gpu specs."""

from __future__ import annotations

import argparse
import importlib
import os
import re
import socket
import subprocess
from dataclasses import dataclass, field, fields
from datetime import datetime
from math import ceil
from pathlib import Path as path
from typing import Any, Optional, TypeVar

import pandas as pd

import config
from utils.amdsmi_interface import (
    amdsmi_ctx,
    get_amdgpu_driver_version,
    get_gpu_compute_partition,
    get_gpu_memory_partition,
    get_gpu_vbios_part_number,
    get_gpu_vram_size,
)
from utils.logger import (
    console_debug,
    console_error,
    console_log,
    console_warning,
    demarcate,
)
from utils.mi_gpu_spec import mi_gpu_specs
from utils.tty import get_table_string
from utils.utils import get_version

T = TypeVar("T")

VERSION_LOC: list[str] = [
    "version",
    "version-dev",
    "version-hip-libraries",
    "version-hiprt",
    "version-hiprt-devel",
    "version-hip-sdk",
    "version-libs",
    "version-utils",
]


def detect_arch(rocminfo_lines: list[str]) -> Optional[tuple[str, int]]:
    supported_gpu_arch = mi_gpu_specs.get_gpu_series_dict()
    unsupported_gpu_arch: set[str] = set()

    for idx1, line_text in enumerate(rocminfo_lines):
        gpu_arch = search(
            r"^\s*Name\s*:\s* ([Gg][Ff][Xx][a-zA-Z0-9]+).*\s*$", line_text
        )
        if not gpu_arch:
            continue

        if gpu_arch in supported_gpu_arch:
            return (gpu_arch, idx1)

        if gpu_arch not in unsupported_gpu_arch:
            unsupported_gpu_arch.add(gpu_arch)
            console_warning(
                "Detected GPU architecture: "
                f"{gpu_arch} is currently NOT supported by the profile mode."
            )

    if unsupported_gpu_arch:
        console_log(f"Supported architectures: {list(supported_gpu_arch.keys())}")

    console_error("Cannot find a supported arch in rocminfo.")


def detect_gpu_chip_id(rocminfo_lines: list[str]) -> Optional[str]:
    chip_id_dict = mi_gpu_specs.get_chip_id_dict()
    unknown_chips: list[str] = []

    for idx, line_text in enumerate(rocminfo_lines):
        chip_id = search(r"^\s*Chip ID\s*:\s* ([0-9]+).*\s*$", line_text)
        if chip_id:
            # Check if this chip ID is valid (known)
            if chip_id in chip_id_dict or int(chip_id) in chip_id_dict:
                return chip_id  # Return first valid chip ID found
            else:
                unknown_chips.append(chip_id)

    # Exhausted all lines - handle the cases where no valid chip was found
    if unknown_chips:
        for chip_id in unknown_chips:
            console_warning(f"Unknown Chip ID(s) detected: {chip_id}")
    else:
        console_warning("No Chip ID detected")

    return None


# Custom decorator to mimic the behavior of kw_only found in Python 3.10
def kw_only(cls: T) -> T:
    def __init__(self: Any, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        for name, value in kwargs.items():
            setattr(self, name, value)

    cls.__init__ = __init__  # type: ignore
    return cls


def generate_machine_specs(
    args: Optional[argparse.Namespace], sysinfo: Optional[dict[str, Any]] = None
) -> MachineSpecs:
    if sysinfo is not None:
        try:
            sysinfo_ver = str(sysinfo["version"])
            version = get_version(config.rocprof_compute_home)["version"]
            curr_ver = version[: version.find(".")]

            if sysinfo_ver != curr_ver:
                console_warning(
                    "Detected mismatch in sysinfo versioning. "
                    "You need to reprofile to update data."
                )

            return MachineSpecs(**sysinfo)
        except KeyError:
            console_error(
                "Detected mismatch in sysinfo versioning. You need to reprofile "
                "to update data."
            )

    # read timestamp info
    now = datetime.now()
    local_now = now.astimezone()
    local_tzname = local_now.tzinfo.tzname(local_now)  # type: ignore
    timestamp = f"{now.strftime('%c')} ({local_tzname})"

    # set specs version
    version = get_version(config.rocprof_compute_home)["version"]
    # NB: Just taking major as specs version.
    # May want to make this more specific in the future
    # version will always follow 'major.minor.patch' format
    specs_version = version[: version.find(".")]

    ##########################################
    ## A. Machine Specs
    ##########################################
    machine_info = extract_machine_info()

    # FIXME: use device
    # Load amd-smi data
    gpu_info = extract_gpu_info()

    ##########################################
    ## B. SoC Specs
    ##########################################
    soc_info = extract_soc_info()

    # Combine all specifications
    with amdsmi_ctx():
        specs = MachineSpecs(
            version=specs_version,
            timestamp=timestamp,
            rocminfo_lines=soc_info["rocminfo_lines"],
            hostname=socket.gethostname(),
            cpu_model=machine_info["cpu_model"],
            sbios=machine_info["sbios"],
            linux_kernel_version=machine_info["linux_kernel_version"],
            amd_gpu_kernel_version=get_amdgpu_driver_version(),
            cpu_memory=machine_info["cpu_memory"],
            gpu_memory=get_gpu_vram_size(),
            linux_distro=machine_info["linux_distro"],
            rocm_version=get_rocm_ver().strip(),
            vbios=gpu_info["vbios"],
            compute_partition=gpu_info["compute_partition"],
            memory_partition=gpu_info["memory_partition"],
            gpu_arch=soc_info["gpu_arch"],
            gpu_chip_id=soc_info["gpu_chip_id"],
        )

    # Load above SoC specs via module import
    try:
        soc_module = importlib.import_module(
            f"rocprof_compute_soc.soc_{specs.gpu_arch}"
        )
        soc_class = getattr(soc_module, f"{specs.gpu_arch}_soc")
        soc_obj = soc_class(args, specs)  # noqa: F841
    except ModuleNotFoundError as e:
        console_error(
            f"Arch {specs.gpu_arch} marked as supported,"
            f"but couldn't find class implementation {e}."
        )

    # Update arch specific specs
    specs.gpu_model = (
        mi_gpu_specs.get_gpu_model(specs.gpu_arch, specs.gpu_chip_id) or ""
    )
    specs.num_xcd = str(
        mi_gpu_specs.get_num_xcds(
            specs.gpu_arch, specs.gpu_model, specs.compute_partition
        )
    )
    specs.total_l2_chan = totall2_banks(
        specs.gpu_arch,
        specs.gpu_model,
        specs.l2_banks,
        specs.compute_partition,
    )
    specs.num_hbm_channels = str(specs.get_hbm_channels())

    return specs


@demarcate
def extract_machine_info() -> dict[str, Any]:
    result: dict[str, Optional[str]] = {
        "cpu_model": None,
        "sbios": None,
        "linux_kernel_version": None,
        "cpu_memory": None,
        "linux_distro": None,
    }

    try:
        cpuinfo = path("/proc/cpuinfo").read_text()
        meminfo = path("/proc/meminfo").read_text()
        version = path("/proc/version").read_text()
        os_release = path("/etc/os-release").read_text()

        result["cpu_model"] = search(r"^model name\s*: (.*?)$", cpuinfo)
        result["sbios"] = (
            path("/sys/class/dmi/id/bios_vendor").read_text().strip()
            + path("/sys/class/dmi/id/bios_version").read_text().strip()
        )
        result["linux_kernel_version"] = search(r"version (\S*)", version)
        result["cpu_memory"] = search(r"MemTotal:\s*(\S*)", meminfo)
        result["linux_distro"] = search(r'PRETTY_NAME="(.*?)"', os_release) or ""

    except OSError as e:
        console_warning(f"Could not read system files: {e}")
    return result


@demarcate
def extract_gpu_info() -> dict[str, Any]:
    result: dict[str, Optional[str]] = {
        "vbios": None,
        "compute_partition": None,
        "memory_partition": None,
    }

    with amdsmi_ctx():
        result["vbios"] = get_gpu_vbios_part_number()
        result["compute_partition"] = get_gpu_compute_partition()
        result["memory_partition"] = get_gpu_memory_partition()

    # Apply defaults and warnings
    if result["compute_partition"] == "N/A" or not result["compute_partition"]:
        console_warning("Cannot detect accelerator partition from amd-smi.")
        console_warning("Applying default accelerator partition: SPX")
        result["compute_partition"] = "SPX"

    if result["memory_partition"] == "N/A" or not result["memory_partition"]:
        console_warning("Cannot detect memory partition from amd-smi.")

    console_debug(
        f"vbios is {result['vbios']}, compute partition is "
        f"{result['compute_partition']}, memory partition is "
        f"{result['memory_partition']}"
    )

    return result


@demarcate
def extract_soc_info() -> dict[str, Any]:
    result: dict[str, Any] = {
        "rocminfo_lines": None,
        "gpu_arch": None,
        "gpu_chip_id": None,
    }

    # Read rocminfo
    rocminfo_full = run(["rocminfo"])
    if rocminfo_full is None:
        return result

    rocminfo_lines = rocminfo_full.split("\n")
    arch_result = detect_arch(rocminfo_lines)

    if arch_result is None:
        return result

    result["gpu_arch"], arch_idx = arch_result
    result["rocminfo_lines"] = rocminfo_lines[
        arch_idx + 1 :
    ]  # update rocminfo for target section
    result["gpu_chip_id"] = detect_gpu_chip_id(rocminfo_lines)

    return result


@kw_only
@dataclass
class MachineSpecs:
    ##########################################
    ## A. Workload / Spec info
    ##########################################

    # these three fields are special in that they're not included
    # when you use (e.g.,) --specs to view the machinespecs, but they
    # _are_ included in profiling/analysis, so we mark them as 'optional'
    # in the metadata to avoid erroring out on missing fields on
    # serialization
    workload_name: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The name of the workload data was collected for.",
            "name": "Workload Name",
            "optional": True,
            "show_in_table": True,
        },
    )
    command: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The command the workload was executed with.",
            "name": "Command",
            "optional": True,
            "show_in_table": True,
        },
    )
    ip_blocks: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The hardware blocks profiling information was collected for.",
            "name": "IP Blocks",
            "optional": True,
        },
    )
    timestamp: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The time (in local system time) when data was collected",
            "name": "Timestamp",
            "show_in_table": True,
        },
    )
    version: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The version of the machine specification file format.",
            "name": "MachineSpecs Version",
            "intable": False,
            "show_in_table": True,
        },
    )
    timestamp: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The time (in local system time) when data was collected",
            "name": "Timestamp",
            "show_in_table": True,
        },
    )
    rocminfo_lines: Optional[list] = field(
        default=None, metadata={"show_in_table": False}
    )
    ##########################################
    ## A. Machine Specs
    ##########################################
    hostname: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The hostname of the machine.",
            "name": "Hostname",
            "show_in_table": True,
        },
    )
    cpu_model: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The model name of the CPU used.",
            "name": "CPU Model",
            "show_in_table": True,
        },
    )
    sbios: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The system management bios version and vendor.",
            "name": "SBIOS",
            "show_in_table": True,
        },
    )
    linux_distro: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The Linux distribution installed on the machine.",
            "name": "Linux Distribution",
            "show_in_table": True,
        },
    )
    linux_kernel_version: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The Linux kernel version running on the machine.",
            "name": "Linux Kernel Version",
            "show_in_table": True,
        },
    )
    amd_gpu_kernel_version: Optional[str] = field(
        default=None,
        metadata={
            "doc": ("The version of the AMDGPU driver installed on the machine."),
            "name": "AMD GPU Kernel Version",
            "show_in_table": True,
        },
    )
    cpu_memory: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The total amount of memory available to the CPU.",
            "unit": "KB",
            "name": "CPU Memory",
            "show_in_table": True,
        },
    )
    gpu_memory: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The total amount of memory available to accelerators/GPUs "
                "in the system."
            ),
            "unit": "KB",
            "name": "GPU Memory",
            "show_in_table": True,
        },
    )
    rocm_version: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The ROCm version used during data-collection.",
            "name": "ROCm Version",
            "show_in_table": True,
        },
    )
    vbios: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The version of the accelerators/GPUs video bios in the system.",
            "name": "VBIOS",
            "show_in_table": True,
        },
    )
    compute_partition: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The compute partitioning mode active on the accelerators/GPUs in the "
                "system (MI300 only)."
            ),
            "name": "Compute Partition",
            "show_in_table": True,
        },
    )
    memory_partition: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The memory partitioning mode active on the accelerators/GPUs in the "
                "system (MI300 only)."
            ),
            "name": "Memory Partition",
            "show_in_table": True,
        },
    )

    ##########################################
    ## B. SoC Specs
    ##########################################
    gpu_series: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The series of the accelerators/GPUs in the system.",
            "name": "GPU Series",
            "show_in_table": True,
        },
    )
    gpu_model: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The product name of the accelerators/GPUs in the system.",
            "name": "GPU Model",
            "show_in_table": True,
        },
    )
    gpu_arch: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The architecture name of the accelerators/GPUs in the system,\n"
            "as used by (e.g.,) the AMDGPU backed of LLVM.",
            "name": "GPU Arch",
            "show_in_table": True,
        },
    )
    gpu_chip_id: Optional[str] = field(
        default=None,
        metadata={
            "doc": "The Chip ID of the accelerators/GPUs in the system.",
            "name": "Chip ID",
            "optional": True,
            "show_in_table": True,
        },
    )
    gpu_l1: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The size of the vL1D cache (per compute-unit) on the "
                "accelerators/GPUs."
            ),
            "name": "GPU L1",
            "unit": "KiB",
            "show_in_table": True,
        },
    )
    gpu_l2: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The size of the vL1D cache (per compute-unit) on the "
                "accelerators/GPUs."
            ),
            "name": "GPU L2",
            "unit": "KiB",
            "show_in_table": True,
        },
    )
    cu_per_gpu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The total number of compute units per accelerator/GPU in the system. "
                "On systems with configurable partitioning, (e.g., MI300) this is "
                "the total number of compute units in a partition."
            ),
            "name": "CU per GPU",
            "show_in_table": True,
        },
    )
    simd_per_cu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number of SIMD processors in a compute unit for the "
                "accelerators/GPUs in the system."
            ),
            "name": "SIMD per CU",
            "show_in_table": True,
        },
    )
    se_per_gpu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number of shader engines on the accelerators/GPUs in the system. "
                "On systems with configurable partitioning, (e.g., MI300) this is "
                "the total number of shader engines in a partition."
            ),
            "name": "SE per GPU",
            "show_in_table": True,
        },
    )
    wave_size: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number work-items in a wavefront on the accelerators/GPUs in "
                "the system."
            ),
            "name": "Wave Size",
            "show_in_table": True,
        },
    )
    workgroup_max_size: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The maximum number of work-items in a workgroup on the "
                "accelerators/GPUs in the system."
            ),
            "name": "Workgroup Max Size",
            "show_in_table": True,
        },
    )
    max_waves_per_cu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The maximum number of wavefronts that can be resident on a "
                "compute unit on the accelerators/GPUs in the system"
            ),
            "name": "Max Waves per CU",
            "show_in_table": True,
        },
    )
    max_sclk: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The maximum engine (compute-unit) clock rate of the "
                "accelerators/GPUs in the system."
            ),
            "name": "Max SCLK",
            "unit": "MHz",
            "show_in_table": True,
        },
    )
    max_mclk: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The maximum memory clock rate of the accelerators/GPUs in the system."
            ),
            "name": "Max MCLK",
            "unit": "MHz",
            "show_in_table": True,
        },
    )
    cur_sclk: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "[RESERVED] The current engine (compute unit) clock rate of the "
                "accelerators/GPUs in the system. Unused."
            ),
            "name": "Cur SCLK",
            "unit": "MHz",
            "show_in_table": True,
        },
    )
    cur_mclk: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "[RESERVED] The current memory clock rate of the accelerators/GPUs "
                "in the system. Unused."
            ),
            "name": "Cur MCLK",
            "unit": "MHz",
            "show_in_table": True,
        },
    )
    l2_banks: Optional[str] = field(
        default=None,
        metadata={
            "show_in_table": True,
        },
    )
    total_l2_chan: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The maximum number of L2 cache channels on the accelerators/GPUs "
                "in the system. On systems with configurable partitioning, "
                "(e.g., MI300) this is the total number of L2 cache channels "
                "in a partition."
            ),
            "name": "Total L2 Channels",
            "show_in_table": True,
        },
    )
    lds_banks_per_cu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number of banks in the LDS for a compute unit on the "
                "accelerators/GPUs in the system."
            ),
            "name": "LDS Banks per CU",
            "show_in_table": True,
        },
    )
    sqc_per_gpu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number of L1I/sL1D caches on the accelerators/GPUs in the "
                "system. On systems with configurable partitioning, (e.g., MI300) "
                "this is the total number of L1I/sL1D caches in a partition."
            ),
            "name": "SQC per GPU",
            "show_in_table": True,
        },
    )
    pipes_per_gpu: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The number of scheduler-pipes on the accelerators/GPUs in the system."
            ),
            "name": "Pipes per GPU",
            "show_in_table": True,
        },
    )
    num_xcd: Optional[str] = field(
        default=None,
        metadata={
            "doc": (
                "The total number of accelerator complex dies in a compute partition "
                "on the accelerators/GPUs in the system. For accelerators without "
                "partitioning (i.e., pre-MI300), this is considered to be one."
            ),
            "name": "Num XCDs",
            "unit": "XCDs",
            "show_in_table": True,
        },
    )
    num_hbm_channels: Optional[str] = field(
        default=None,
        metadata={
            "doc": "Number of HBM channels",
            "name": "HBM channels",
            "show_in_table": True,
        },
    )

    def get_hbm_channels(self) -> Optional[str]:
        if self.memory_partition and self.memory_partition.lower().startswith("nps"):
            hbmchannels = 128
            if self.memory_partition.lower() == "nps4":
                hbmchannels //= 4
            elif self.memory_partition.lower() == "nps8":
                hbmchannels //= 8
            return str(hbmchannels)
        else:
            return self.total_l2_chan

    def get_class_members(self) -> pd.DataFrame:
        data = {}
        missing_required_fields = []

        for class_field in fields(self):
            if not class_field.metadata.get("show_in_table", True):
                continue

            name = class_field.name
            value = getattr(self, name)
            data[name] = value

            # Check for missing required fields
            if value is None and not class_field.metadata.get("optional", False):
                missing_required_fields.append(name)

        # Handle warnings after processing all fields
        if missing_required_fields:
            for field_name in missing_required_fields:
                console_warning(
                    f"Incomplete class definition for {self.gpu_arch}. "
                    f"Expecting populated {field_name} but detected None."
                )
            console_warning(f"Missing specs fields for {self.gpu_arch}")

        return pd.DataFrame(data, index=[0])

    def __repr__(self) -> str:
        topstr = (
            "Machine Specifications: describing the state of the machine that "
            "ROCm Compute Profiler data was collected on.\n"
        )
        data = []
        for class_field in fields(self):
            name = class_field.name
            if class_field.metadata.get("show_in_table", True):
                _data = {}
                value = getattr(self, name)
                if class_field.metadata:
                    # check out of table before any re-naming for pretty-printing
                    if (
                        "intable" in class_field.metadata
                        and not class_field.metadata["intable"]
                    ):
                        if name == "version":
                            topstr += f"Output version: {value}\n"
                        else:
                            console_error(
                                f"Unknown out of table printing field: {name}"
                            )
                        continue
                    if "name" in class_field.metadata:
                        name = class_field.metadata["name"]
                    if "unit" in class_field.metadata:
                        _data["Unit"] = class_field.metadata["unit"]
                    if "doc" in class_field.metadata:
                        _data["Description"] = class_field.metadata["doc"]
                _data["Spec"] = name
                _data["Value"] = value
                data.append(_data)
        df = pd.DataFrame(data)
        columns = ["Spec", "Value"]
        if "Description" in df.columns:
            columns += ["Description"]
        if "Unit" in df.columns:
            columns += ["Unit"]
        df = df[columns]
        df = df.fillna("")
        return topstr + get_table_string(df, transpose=False, decimal=2)


def get_rocm_ver() -> str:
    # Check for version files in ROCm installation
    rocm_base_path = path(os.getenv("ROCM_PATH", "/opt/rocm"))

    for version_file_name in VERSION_LOC:
        version_file_path = rocm_base_path / ".info" / version_file_name
        if version_file_path.exists():
            return version_file_path.read_text().strip()

    # Fallback to environment variable
    ROCM_VER_USER = os.getenv("ROCM_VER")
    if ROCM_VER_USER:
        console_log(
            "profiling",
            "Overriding missing ROCm version detection with "
            f"ROCM_VER = {ROCM_VER_USER}",
        )
        return ROCM_VER_USER

    # No version found - log error and return empty string
    console_warning("Unable to detect a complete local ROCm installation.")
    console_warning(
        f"The expected {rocm_base_path}/.info/ versioning directory is missing."
    )
    console_error("Ensure you have valid ROCm installation.", exit=False)
    return ""


def run(cmd: list[str], exit_on_error: bool = False) -> str:
    try:
        p = subprocess.run(cmd, capture_output=True)
    except FileNotFoundError as e:
        console_error(
            f"Unable to parse specs. Can't find ROCm asset: {e.filename}\n"
            'Try passing a path to an existing workload results in "analyze" mode.'
        )

    if exit_on_error and p.returncode != 0:  # type: ignore
        console_error(f"Command {cmd} failed with non-zero exit code")
    return p.stdout.decode("utf-8")  # type: ignore


def search(pattern: str, string: str) -> Optional[str]:
    m = re.search(pattern, string, re.MULTILINE)
    if m is not None:
        return m.group(1)
    return None


def total_sqc(archname: str, num_compute_units: str, num_shader_engines: str) -> int:
    cu_per_se = float(num_compute_units) / float(num_shader_engines)
    sq_per_se = cu_per_se / 2.0
    if archname.lower() in ["mi50", "mi100"]:
        sq_per_se = cu_per_se / 3
    sq_per_se = ceil(sq_per_se)
    return int(sq_per_se) * int(num_shader_engines)


def totall2_banks(
    gpu_arch: Optional[str] = None,
    gpu_model: Optional[str] = None,
    L2banks: Optional[str] = None,
    compute_partition: Optional[str] = None,
) -> Optional[str]:
    xcd_count = mi_gpu_specs.get_num_xcds(gpu_arch, gpu_model, compute_partition)

    # TODO: MachineSpecs and OmniSoC mspec should converge...
    if L2banks is not None and xcd_count is not None:
        return str(int(L2banks) * int(xcd_count))
    return None


if __name__ == "__main__":
    specs = generate_machine_specs(None, None)
    if specs:
        print(specs)
    else:
        console_error("specs", "Failed to generate machine specifications", exit=False)



##############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

import csv
import sqlite3
from contextlib import closing
from typing import Any

import pandas as pd

from utils.logger import console_error

# From schema definition in source/share/rocprofiler-sdk-rocpd/data_views.sql
# in rocprofiler-sdk repository
COUNTERS_COLLECTION_QUERY = """
SELECT
    agent_id as GPU_ID,
    dispatch_id as Dispatch_ID,
    pid as PID,
    grid_size as Grid_Size,
    workgroup_size as Workgroup_Size,
    lds_block_size as LDS_Per_Workgroup,
    scratch_size as Scratch_Per_Workitem,
    vgpr_count as Arch_VGPR,
    accum_vgpr_count as Accum_VGPR,
    sgpr_count as SGPR,
    kernel_name as Kernel_Name,
    start as Start_Timestamp,
    end as End_Timestamp,
    kernel_id as Kernel_ID,
    counter_name as Counter_Name,
    value as Counter_Value
FROM counters_collection
"""
ROCPD_PMC_EVENT_TABLE_NAME_PREFIX = "rocpd_pmc_event_"
TABLE_NAME_PREFIX_QUERY = (
    "SELECT name FROM sqlite_master WHERE type='table' "
    "AND name LIKE '{table_name_prefix}%'"
)
INSERT_QUERY = "INSERT INTO {table_name} ({columns}) VALUES ({placeholders})"


def convert_dbs_to_csv(
    db_paths: list[str],
    csv_file_path: str,
) -> None:
    """
    Read rocpd databases and write to CSV file
    """
    # Read counters_collection view from the databases and write to CSV
    try:
        with open(csv_file_path, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            header_written = False
            for db_path in db_paths:
                with closing(sqlite3.connect(db_path)) as conn:
                    with closing(conn.execute(COUNTERS_COLLECTION_QUERY)) as cursor:
                        if not header_written:
                            writer.writerow([
                                description[0] for description in cursor.description
                            ])
                            header_written = True
                        for row in cursor:
                            writer.writerow(row)
    except OSError as e:
        console_error(f"Database error while converting to CSV: {e}")
    except Exception as e:
        console_error(f"Unexpected error converting database to CSV: {e}")


def process_rocpd_csv(df: pd.DataFrame) -> pd.DataFrame:
    """
    Merge counters across unique dispatches from the
    input dataframe and return processed dataframe.
    """
    if df.empty:
        return df

    data: list[dict[str, Any]] = []

    # Group by unique kernel and merge into a single row
    for _, group_df in df.groupby([
        "Dispatch_ID",
        "Kernel_Name",
        "Grid_Size",
        "Workgroup_Size",
        "LDS_Per_Workgroup",
    ]):
        row = {
            "GPU_ID": group_df["GPU_ID"].iloc[0],
            "Grid_Size": group_df["Grid_Size"].iloc[0],
            "Workgroup_Size": group_df["Workgroup_Size"].iloc[0],
            "LDS_Per_Workgroup": group_df["LDS_Per_Workgroup"].iloc[0],
            "Scratch_Per_Workitem": group_df["Scratch_Per_Workitem"].iloc[0],
            "Arch_VGPR": group_df["Arch_VGPR"].iloc[0],
            "Accum_VGPR": group_df["Accum_VGPR"].iloc[0],
            "SGPR": group_df["SGPR"].iloc[0],
            "Kernel_Name": group_df["Kernel_Name"].iloc[0],
            "Kernel_ID": group_df["Kernel_ID"].iloc[0],
            "Start_Timestamp": group_df["Start_Timestamp"].iloc[0],
            "End_Timestamp": group_df["End_Timestamp"].iloc[0],
        }
        # Each counter will become its own column
        row.update(dict(zip(group_df["Counter_Name"], group_df["Counter_Value"])))
        data.append(row)
    df = pd.DataFrame(data)
    # Rank GPU IDs, map lowest number to 0, next to 1, etc.
    df["GPU_ID"] = df["GPU_ID"].rank(method="dense").astype(int) - 1
    # Reset dispatch IDs
    df["Dispatch_ID"] = range(len(df))
    return df


def update_rocpd_pmc_events(counter_info: pd.DataFrame, rocpd_db_path: str) -> None:
    """Update pmc_event table in the given rocpd database path"""
    try:
        with closing(sqlite3.connect(rocpd_db_path)) as conn:
            # Get pmc_event table name
            with closing(
                conn.execute(
                    TABLE_NAME_PREFIX_QUERY.format(
                        table_name_prefix=ROCPD_PMC_EVENT_TABLE_NAME_PREFIX
                    )
                )
            ) as cursor:
                table_name = cursor.fetchone()
            if table_name is None:
                console_error("No pmc_event table found in the rocpd database")
            table_name = table_name[0]

            # get pmc_event table data
            guid = table_name[len(ROCPD_PMC_EVENT_TABLE_NAME_PREFIX) :].replace(
                "_", "-"
            )
            columns = ("guid", "event_id", "pmc_id", "value")
            values = list(
                zip(
                    # guid
                    [guid] * len(counter_info),
                    # event_id
                    counter_info["dispatch_id"],
                    # pmc_id
                    counter_info["counter_id"],
                    # value
                    counter_info["counter_value"],
                )
            )

            # insert into pmc_event table
            with conn:
                placeholders = ", ".join(["?"] * len(columns))
                conn.executemany(
                    INSERT_QUERY.format(
                        table_name=table_name,
                        columns=", ".join(columns),
                        placeholders=placeholders,
                    ),
                    values,
                )
    except OSError as e:
        console_error(f"Database error while updating pmc_event table: {e}")
    except Exception as e:
        console_error(f"Unexpected error updating pmc_event table: {e}")



##############################################################################
# MIT License
#
# Copyright (c) 2021 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

from typing import Any

import pandas as pd
import plotly.express as px  # type: ignore
from dash import dash_table  # type: ignore

from utils.logger import console_error

pd.set_option(
    "mode.chained_assignment", None
)  # ignore SettingWithCopyWarning pandas warning


####################
# GRAPHICAL ELEMENTS
####################
def multi_bar_chart(
    table_id: int, display_df: pd.DataFrame
) -> dict[str, dict[str, Any]]:
    nested_bar: dict[str, dict[str, Any]] = {}
    if table_id == 1604:
        for _, row in display_df.iterrows():
            coherency = row["Coherency"]
            if coherency not in nested_bar:
                nested_bar[coherency] = {}
            nested_bar[coherency][row["Xfer"]] = row["Avg"]
    elif table_id == 1705:  # L2 - Fabric Interface Stalls
        for _, row in display_df.iterrows():
            transaction = row["Transaction"]
            if transaction not in nested_bar:
                nested_bar[transaction] = {}
            nested_bar[transaction][row["Type"]] = row["Avg"]

    return nested_bar


def create_instruction_mix_bar_chart(display_df: pd.DataFrame, df_unit: str) -> px.bar:
    display_df = display_df.copy()
    display_df["Avg"] = display_df["Avg"].apply(lambda x: int(x) if x != "" else 0)

    return px.bar(
        display_df,
        x="Avg",
        y="Metric",
        color="Avg",
        labels={"Avg": f"# of {df_unit.lower()}"},
        height=400,
        orientation="h",
    )


def create_multi_bar_charts(
    display_df: pd.DataFrame, table_id: int, df_unit: str
) -> list[px.bar]:
    display_df = display_df.copy()
    display_df["Avg"] = display_df["Avg"].apply(lambda x: int(x) if x != "" else 0)

    nested_bar = multi_bar_chart(table_id, display_df)
    charts = []

    for group, metric in nested_bar.items():
        chart = px.bar(
            title=group,
            x=list(metric.values()),
            y=list(metric.keys()),
            labels={"x": df_unit, "y": ""},
            text=list(metric.values()),
            orientation="h",
            height=200,
        )
        chart.update_xaxes(showgrid=False, rangemode="nonnegative")
        chart.update_yaxes(showgrid=False)
        chart.update_layout(title_x=0.5)
        charts.append(chart)

    return charts


def create_sol_charts(display_df: pd.DataFrame, table_id: int) -> list[px.bar]:
    display_df = display_df.copy()
    display_df["Avg"] = display_df["Avg"].apply(lambda x: float(x) if x != "" else 0.0)

    charts = []

    if table_id == 1701:
        # Special layout for L2 Cache SOL
        pct_data = display_df[display_df["Unit"] == "Pct"]
        charts.append(
            px.bar(
                pct_data,
                x="Avg",
                y="Metric",
                color="Avg",
                range_color=[0, 100],
                labels={"Avg": "%"},
                height=220,
                orientation="h",
            ).update_xaxes(range=[0, 110], ticks="inside", title="%")
        )

        # HBM Bandwidth chart
        hbm_row = display_df[display_df["Metric"] == "HBM Bandwidth"]
        if not hbm_row.empty:
            hbm_bw = float(hbm_row["Avg"].iloc[0])
            gb_data = display_df[display_df["Unit"] == "Gb/s"]
            charts.append(
                px.bar(
                    gb_data,
                    x="Avg",
                    y="Metric",
                    color="Avg",
                    range_color=[0, hbm_bw],
                    labels={"Avg": "GB/s"},
                    height=220,
                    orientation="h",
                ).update_xaxes(range=[0, hbm_bw])
            )

    elif table_id == 1101:
        # Special formatting reference 'Pct of Peak' value
        display_df["Pct of Peak"] = display_df["Pct of Peak"].apply(
            lambda x: float(x) if x != "" else 0.0
        )
        charts.append(
            px.bar(
                display_df,
                x="Pct of Peak",
                y="Metric",
                color="Pct of Peak",
                range_color=[0, 100],
                labels={"Avg": "%"},
                height=400,
                orientation="h",
            ).update_xaxes(range=[0, 110])
        )
    else:
        charts.append(
            px.bar(
                display_df,
                x="Avg",
                y="Metric",
                color="Avg",
                range_color=[0, 100],
                labels={"Avg": "%"},
                height=400,
                orientation="h",
            ).update_xaxes(range=[0, 110])
        )

    return charts


def build_bar_chart(
    display_df: pd.DataFrame,
    table_config: dict[str, Any],
    barchart_elements: dict[str, Any],
) -> list:
    """
    Read data into a bar chart. ID will determine which subtype of barchart.
    """
    table_id = table_config["id"]
    charts: list[px.bar] = []

    # Get unit from first row if available
    df_unit = display_df["Unit"].iloc[0] if "Unit" in display_df.columns else ""

    # Instruction Mix bar chart
    if table_id in barchart_elements["instr_mix"]:
        charts.append(create_instruction_mix_bar_chart(display_df, df_unit))

    # Multi bar chart
    elif table_id in barchart_elements["multi_bar"]:
        charts.extend(create_multi_bar_charts(display_df, table_id, df_unit))

    # Speed-of-light bar chart
    elif table_id in barchart_elements["sol"]:
        charts.extend(create_sol_charts(display_df, table_id))

    else:
        console_error(
            f"Table id {table_id}. Cannot determine barchart type.", exit=False
        )
        return []

    # Apply consistent styling to all charts
    for fig in charts:
        fig.update_layout(
            margin=dict(l=50, r=50, b=50, t=50, pad=4),
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
            font={"color": "#ffffff"},
        )

    return charts


def get_dark_mode_styles() -> tuple[
    dict[str, Any], dict[str, Any], list[dict[str, Any]]
]:
    style_header = {
        "backgroundColor": "rgb(30, 30, 30)",
        "color": "white",
        "fontWeight": "bold",
    }

    style_data = {
        "backgroundColor": "rgb(50, 50, 50)",
        "color": "white",
        "whiteSpace": "normal",
        "height": "auto",
    }

    style_data_conditional = [
        {"if": {"row_index": "odd"}, "backgroundColor": "rgb(60, 60, 60)"}
    ]

    return style_header, style_data, style_data_conditional


def build_table_chart(
    display_df: pd.DataFrame,
    table_config: dict[str, Any],
    original_df: pd.DataFrame,
    display_columns: list[str],
    comparable_columns: list[str],
    decimal: int,
) -> list[dash_table.DataTable]:
    """
    Read data into a DashTable
    """
    d_figs = []

    # build comlumns/header with formatting
    formatted_columns = []
    for col in display_df.columns:
        col_lower = str(col).lower()
        if col_lower in {"pct", "pop", "percentage"}:
            formatted_columns.append({
                "id": col,
                "name": col,
                "type": "numeric",
                "format": {"specifier": f".{decimal}f"},
            })
        elif col in comparable_columns:
            formatted_columns.append({
                "id": col,
                "name": col,
                "type": "numeric",
                "format": {"specifier": f".{decimal}f"},
            })
        else:
            formatted_columns.append({"id": col, "name": col, "type": "text"})

    # tooltip shows only on the 1st col for now if 'Metric Description' available
    table_tooltip = (
        [
            {
                column: {
                    "value": (
                        str(row["Description"])
                        if column == display_columns[0] and row["Description"]
                        else ""
                    ),
                    "type": "markdown",
                }
                for column in row.keys()
            }
            for row in original_df.to_dict("records")
        ]
        if "Description" in original_df.columns.values.tolist()
        else None
    )

    # Get styling based on dark mode
    style_header, style_data, style_data_conditional = get_dark_mode_styles()

    # build data table with columns, tooltip, df and other properties
    d_t = dash_table.DataTable(
        id=str(table_config["id"]),
        sort_action="native",
        sort_mode="multi",
        columns=formatted_columns,
        tooltip_data=table_tooltip,
        # left-aligning the text of the 1st col
        style_cell_conditional=[
            {"if": {"column_id": display_columns[0]}, "textAlign": "left"}
        ],
        # style cell
        style_cell={"maxWidth": "500px"},
        # display style
        style_header=style_header,
        style_data=style_data,
        style_data_conditional=style_data_conditional,
        # the df to display
        data=display_df.to_dict("records"),
    )
    d_figs.append(d_t)
    return d_figs



#!/usr/bin/env python3
##############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

import argparse
import shutil
import subprocess
import sys
import time
from pathlib import Path

SCRIPT_DIR = Path(__file__).resolve().parent
# .../rocprofiler-compute/tools/config_management

REPO_ROOT = SCRIPT_DIR.parents[1]
# .../rocprofiler-compute

TOOLS_DIR = SCRIPT_DIR

SOC_ROOT = REPO_ROOT / "src" / "rocprof_compute_soc"
ANALYSIS_CONFIGS = SOC_ROOT / "analysis_configs"

TEMPLATE_FILE = ANALYSIS_CONFIGS / "gfx9_config_template.yaml"
HASH_JSON = REPO_ROOT / "src" / "utils" / ".config_hashes.json"
BACKUP_DIR = SCRIPT_DIR / "backups"

PYTHON = sys.executable

VERIFY_SCRIPT = TOOLS_DIR / "verify_against_config_template.py"
PARSE_TEMPLATE_SCRIPT = TOOLS_DIR / "parse_config_template.py"
GENERATE_DELTAS_SCRIPT = TOOLS_DIR / "generate_config_deltas.py"
APPLY_DELTAS_SCRIPT = TOOLS_DIR / "apply_config_deltas.py"
HASH_CHECKER_SCRIPT = REPO_ROOT / "src" / "utils" / "hash_checker.py"
HASH_MANAGER_SCRIPT = TOOLS_DIR / "hash_manager.py"


def run(cmd):
    print("\n$", " ".join(str(c) for c in cmd))
    return subprocess.run(cmd, cwd=str(REPO_ROOT)).returncode


def fatal(msg):
    print(f"\nFATAL: {msg}")
    sys.exit(1)


def confirm(prompt):
    ans = input(f"{prompt} [y/N]: ").strip().lower()
    return ans in ("y", "yes")


def backup(paths):
    BACKUP_DIR.mkdir(exist_ok=True)
    backup_path = BACKUP_DIR / f"backup_{int(time.time())}"
    backup_path.mkdir()

    for p in paths:
        if not p.exists():
            continue
        dest = backup_path / p.name
        if p.is_dir():
            shutil.copytree(p, dest)
        else:
            shutil.copy2(p, dest)

    print(f"\nBackup created at {backup_path}")
    return backup_path


def restore(backup_path, paths):
    print("\nRestoring from backup...")
    for p in paths:
        src = backup_path / p.name
        if not src.exists():
            continue
        if p.exists():
            if p.is_dir():
                shutil.rmtree(p)
            else:
                p.unlink()
        if src.is_dir():
            shutil.copytree(src, p)
        else:
            shutil.copy2(src, p)
    print("Restore complete.")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ci", action="store_true")
    parser.add_argument("--hash-only", action="store_true")
    parser.add_argument("--validate-only", action="store_true")
    parser.add_argument("--edit-existing", action="store_true")
    parser.add_argument("--promote", metavar="ARCH")
    args = parser.parse_args()

    # --------------------------------------------------------
    # CI / HASH-ONLY PATH (non-mutating)
    # --------------------------------------------------------
    if args.ci or args.hash_only:
        if not HASH_CHECKER_SCRIPT.exists():
            fatal("hash_checker.py not found")
        sys.exit(run([PYTHON, HASH_CHECKER_SCRIPT]))

    # --------------------------------------------------------
    # HARD PREFLIGHT (STRUCTURAL VALIDATION) for all non-hash paths
    # --------------------------------------------------------
    if not VERIFY_SCRIPT.exists():
        fatal("verify_against_config_template.py not found")

    rc = run([PYTHON, VERIFY_SCRIPT, ANALYSIS_CONFIGS, TEMPLATE_FILE])
    if rc != 0:
        fatal("Template / architecture verification failed")

    # --------------------------------------------------------
    # VALIDATE-ONLY
    # --------------------------------------------------------
    if args.validate_only:
        print("\nValidation successful.")
        sys.exit(0)

    # --------------------------------------------------------
    # EDIT EXISTING ARCHITECTURE (helpers only; no template/hash updates)
    # --------------------------------------------------------
    if args.edit_existing:
        print("\nEdit existing architecture mode.")

        choice = input(
            "\nChoose:\n  1) Generate delta\n  2) Apply delta\n  3) Exit\nSelect: "
        ).strip()

        if choice == "1":
            base = input("Base arch dir (absolute or relative to repo root): ").strip()
            new = input("New  arch dir: ").strip()
            out = input("Output delta yaml: ").strip()
            sys.exit(run([PYTHON, GENERATE_DELTAS_SCRIPT, base, new, out]))

        if choice == "2":
            base = input("Base arch dir: ").strip()
            delta = input("Delta yaml: ").strip()
            out = input("Output dir: ").strip()

            rc = run([PYTHON, APPLY_DELTAS_SCRIPT, base, delta, out])
            if rc != 0:
                sys.exit(rc)

            # Re-verify after apply
            rc = run([PYTHON, VERIFY_SCRIPT, ANALYSIS_CONFIGS, TEMPLATE_FILE])
            sys.exit(rc)

        sys.exit(0)

    # --------------------------------------------------------
    # PROMOTE NEW LATEST ARCHITECTURE (mutating, with rollback)
    # --------------------------------------------------------
    if args.promote:
        new_latest = args.promote
        new_arch_dir = ANALYSIS_CONFIGS / new_latest

        if not new_arch_dir.is_dir():
            fatal(f"Architecture directory not found: {new_arch_dir}")

        if not confirm(
            f"Promote {new_latest} to latest? "
            "This will update template, regenerate deltas, and update hashes."
        ):
            sys.exit(0)

        # Back up the things we mutate
        backup_path = backup([ANALYSIS_CONFIGS, TEMPLATE_FILE, HASH_JSON])

        try:
            # 1) Update template
            if not PARSE_TEMPLATE_SCRIPT.exists():
                raise RuntimeError("parse_config_template.py not found")

            rc = run([
                PYTHON,
                PARSE_TEMPLATE_SCRIPT,
                new_arch_dir,
                TEMPLATE_FILE,
                "--latest-arch",
                new_latest,
            ])
            if rc != 0:
                raise RuntimeError("Failed to update template")

            # 2) Regenerate deltas for all other archs
            for arch_dir in sorted(ANALYSIS_CONFIGS.iterdir()):
                if not arch_dir.is_dir():
                    continue
                if arch_dir.name == new_latest:
                    continue

                delta_dir = arch_dir / "config_delta"
                delta_dir.mkdir(exist_ok=True)
                out_delta = delta_dir / f"{new_latest}_diff.yaml"

                rc = run([
                    PYTHON,
                    GENERATE_DELTAS_SCRIPT,
                    new_arch_dir,
                    arch_dir,
                    out_delta,
                ])
                if rc != 0:
                    raise RuntimeError(f"Delta generation failed for {arch_dir.name}")

                for f in delta_dir.glob("*_diff.yaml"):
                    if f.name != f"{new_latest}_diff.yaml":
                        f.unlink()

            # 3) Re-verify everything against updated template
            rc = run([PYTHON, VERIFY_SCRIPT, ANALYSIS_CONFIGS, TEMPLATE_FILE])
            if rc != 0:
                raise RuntimeError("Post-promotion verification failed")

            # 4) Now update the hash DB to the new steady state.
            #    Promotion touched many delta files, so compute-all is the safest.
            if not HASH_MANAGER_SCRIPT.exists():
                raise RuntimeError("hash_manager.py not found")

            rc = run([
                PYTHON,
                HASH_MANAGER_SCRIPT,
                "--compute-all",
                ANALYSIS_CONFIGS,
                HASH_JSON,
            ])
            if rc != 0:
                raise RuntimeError("Hash DB update failed (--compute-all)")

            # 5) run hash_checker
            rc = run([PYTHON, HASH_CHECKER_SCRIPT])
            if rc != 0:
                raise RuntimeError(
                    "Hash consistency check failed (after hash DB update)"
                )

            print(f"\nSUCCESS: {new_latest} promoted to latest.")
            sys.exit(0)

        except Exception as e:
            print(f"\nERROR: {e}")
            restore(backup_path, [ANALYSIS_CONFIGS, TEMPLATE_FILE, HASH_JSON])
            sys.exit(1)

    # --------------------------------------------------------
    # NO INTENT PROVIDED
    # --------------------------------------------------------
    print(
        "\nNo workflow selected.\n"
        "Use one of:\n"
        "  --validate-only\n"
        "  --edit-existing\n"
        "  --promote gfxXYZ\n"
        "  --hash-only / --ci\n"
    )
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
##############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################

"""
parse_config_template.py

Parse panel configuration based on YAML files for an architecture and, optionally,
generate a lightweight template describing panel IDs, titles, aliases, and
data-source ordering.

Usage
-----
Generate a template from an architecture directory:

    python tools/config_management/parse_config_template.py \
        analysis_configs/gfx950 \
        analysis_configs/config_template.yaml \
        --latest-arch gfx950

Inspect an architecture (no template written):

    python tools/config_management/parse_config_template.py \
        analysis_configs/gfx950
"""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import Any, Optional

PROJECT_ROOT = Path(__file__).resolve().parents[1]
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from config_management import utils_ruamel as cm_utils  # noqa: E402


def normalize_panel_id(panel_id: Optional[int]) -> Optional[int]:
    """Normalize panel ID by dividing by 100 if needed."""
    if panel_id is None:
        return None
    return panel_id // 100 if panel_id >= 100 else panel_id


def normalize_table_id(table_id: Optional[int]) -> Optional[int]:
    """Normalize table ID using modulo 100."""
    if table_id is None:
        return None
    return table_id % 100


def parse_panel_config(yaml_file: Path) -> Optional[dict[str, Any]]:
    """
    Parse a single panel YAML file and extract template-relevant info.

    Returns a dict with:
      - file: panel filename (without leading numeric prefix)
      - panel_id: normalized panel id (id // 100 when >= 100)
      - panel_title: Panel Config.title
      - panel_alias: Panel Config.alias (optional)
      - data_sources: ordered list of
            {type: <key>, id: <normalized table id>, title: <title>}
    or None if the file does not contain a valid Panel Config or fails basic checks.
    """
    data = cm_utils.load_yaml(yaml_file)
    panel_config = data.get("Panel Config")
    if not isinstance(panel_config, dict):
        print(f"WARNING: {yaml_file} has no valid 'Panel Config' mapping, skipping.")
        return None

    # Enforce presence of core panel-level keys
    missing_keys: list[str] = []
    for key in ("id", "title", "data source", "metrics_description"):
        if key not in panel_config:
            missing_keys.append(key)

    if missing_keys:
        missing_str = ", ".join(missing_keys)
        print(
            f"ERROR: {yaml_file} is missing required Panel Config keys: {missing_str}"
        )
        return None

    filename = (
        yaml_file.name.split("_", 1)[1] if "_" in yaml_file.name else yaml_file.name
    )

    raw_panel_id = panel_config.get("id")
    if not isinstance(raw_panel_id, int):
        print(
            f"ERROR: {yaml_file} has non-integer or missing Panel Config.id "
            f"({raw_panel_id!r})"
        )
        return None

    panel_id = normalize_panel_id(raw_panel_id)

    # Extract and normalize data sources
    data_sources: list[dict[str, Any]] = []
    ds_list = panel_config.get("data source", [])
    if not isinstance(ds_list, list):
        print(
            f"ERROR: {yaml_file} has non-list 'data source' field "
            f"({type(ds_list).__name__})"
        )
        return None

    for ds in ds_list:
        if not isinstance(ds, dict):
            print(f"WARNING: {yaml_file} has non-dict data source entry: {ds!r}")
            continue
        for key, value in ds.items():
            if isinstance(value, dict) and "id" in value and "title" in value:
                norm_id = normalize_table_id(value["id"])
                data_sources.append({
                    "type": key,
                    "id": norm_id,
                    "title": value["title"],
                })

    return {
        "file": filename,
        "panel_id": panel_id,
        "panel_title": panel_config.get("title"),
        "panel_alias": panel_config.get("alias"),
        "data_sources": data_sources,
    }


def build_template_from_directory(
    directory: Path,
    existing_panels_by_id: Optional[dict[int, dict]],
) -> list[dict]:
    panels: list[dict] = []
    errors = 0

    for yaml_file in sorted(directory.glob("*.yaml")):
        info = parse_panel_config(yaml_file)
        if info is None:
            errors += 1
            continue

        panel_id = info.get("panel_id")

        if (
            existing_panels_by_id
            and panel_id is not None
            and panel_id in existing_panels_by_id
        ):
            old_panel = existing_panels_by_id[panel_id]

            # Preserve panel_alias unless explicitly set by panel YAML
            if info.get("panel_alias") is None and "panel_alias" in old_panel:
                info["panel_alias"] = old_panel["panel_alias"]

        panels.append(info)

    # Deterministic ordering for stable templates
    panels.sort(key=lambda p: (p["panel_id"], p["file"]))

    if errors:
        print(
            f"\nEncountered {errors} panel file(s) with structural errors "
            "while building template."
        )

    return panels


def main() -> None:
    parser = argparse.ArgumentParser(
        description=(
            "Parse panel YAML files for an architecture and optionally generate "
            "a config_template-style YAML describing panel IDs and data sources."
        )
    )
    parser.add_argument("directory", help="Directory containing panel YAML files")
    parser.add_argument(
        "output",
        nargs="?",
        help="Output YAML file (optional). If omitted, only a summary is printed.",
    )
    parser.add_argument(
        "--latest-arch",
        help=(
            "Specify this architecture as latest (adds 'latest_arch' metadata "
            "to the generated template). Only used when an output file is given."
        ),
    )
    args = parser.parse_args()

    directory = Path(args.directory)
    if not directory.is_dir():
        print(f"Error: '{args.directory}' is not a valid directory")
        sys.exit(1)

    existing_template = None
    if args.output and Path(args.output).exists():
        existing_template = cm_utils.load_yaml(Path(args.output))

    existing_panels_by_id = {}
    if existing_template:
        for p in existing_template.get("panels", []):
            pid = p.get("panel_id")
            if pid is not None:
                existing_panels_by_id[pid] = p

    panels = build_template_from_directory(
        directory,
        existing_panels_by_id=existing_panels_by_id if args.output else None,
    )

    if not panels:
        print("No valid panel YAML files found; nothing to do.")
        sys.exit(1)

    # Always show a human-readable summary.
    print(f"Found {len(panels)} panel(s) in {directory}:")
    for panel in panels:
        print(f"\nFile: {panel['file']}")
        print(f"Panel ID: {panel['panel_id']}")
        print(f"Panel Title: {panel['panel_title']}")
        if panel["panel_alias"]:
            print(f"Panel Alias: {panel['panel_alias']}")
        print(f"\nData Sources ({len(panel['data_sources'])}):")
        for ds in panel["data_sources"]:
            print(f"  - {ds['type']}: {ds['id']} - {ds['title']}")

    # Optionally write a template YAML.
    if args.output:
        output_data: Any = {"panels": panels}
        if args.latest_arch:
            output_data = {"latest_arch": args.latest_arch, "panels": panels}
        cm_utils.save_yaml(output_data, Path(args.output))
        print(f"\nTemplate saved to: {args.output}")


if __name__ == "__main__":
    main()
